# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.deprecated_middleware import deprecated_asgi
from contrast.patches.middleware.django import initialize_django

ASGIMiddleware = deprecated_asgi("django")


class DjangoASGIMiddleware(ASGIMiddleware):
    def __init__(self, *args, **kwargs):
        initialize_django()
        super().__init__(*args, **kwargs)



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.asgi.middleware import ASGIMiddleware

FastApiMiddleware = ASGIMiddleware



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from functools import cached_property
from typing import Optional, Callable

from contrast.agent import scope

from contrast.agent.exclusions import Exclusions
from contrast.agent.request import Request
from contrast.agent.settings import Settings
from contrast.api.route_coverage import Route
from contrast_vendor import structlog as logging
from contrast.utils.digest_utils import Digest
from contrast.utils.decorators import fail_quietly
from contrast.reporting.request_masker import RequestMasker

logger = logging.getLogger("contrast")


class RequestContext:
    @scope.contrast_scope()
    def __init__(
        self,
        environ,
        assess_enabled: bool,
        exclusions: Optional[Exclusions] = None,
        request_data_masker: Optional[RequestMasker] = None,
    ):
        self.request = Request(environ)
        self.response = None

        # This contains any Observed Library Usage seen during the life of this request.
        self.observed_library_usage = None

        self.exclusions = exclusions

        self.request_data_masker = None

        if request_data_masker:
            self.request_data_masker = RequestMasker.new_request_masker(
                request_data_masker.mask_rules
            )

        # This contains input exclusions that should always apply regardless of rule on this request
        self.input_exclusions = None

        # This contains exclusions that only apply at trigger time based on the source event
        self.input_exclusions_trigger_time = None

        self.excluded_assess_rules = []
        self.excluded_protect_rules = []

        # For protect: store attacks made during a request to report at the end
        self.attacks = []

        # For assess: store findings made during a request to report at the end
        self.findings = []

        self.user_input_analysis = []
        self.response = None
        self.do_not_track = False

        self.view_func: Optional[Callable] = None
        self.view_func_str: Optional[str] = None

        self.observed_route = Route("", "", "")
        self.database_info = []

        self.assess_enabled = assess_enabled
        if self.assess_enabled:
            self.source_count = 0
            self.propagation_count = 0

            self.max_sources_logged = False
            self.max_propagators_logged = False

    @cached_property
    @fail_quietly("Unable to compute request hash")
    def hash(self):
        """
        Generates the hash checksum for the request. Converts the request method, uri,
        param names and content length to CRC checksum and returns string representation

        :return: str
        """
        hasher = Digest()

        hasher.update(self.request.method)
        hasher.update(self.request.get_normalized_uri())
        for key in self.request.params:
            hasher.update(key)

        hasher.update(self.request.content_length)
        return hasher.finish()

    def stop_source_creation(self, source_type, source_name):
        """
        Compare `source_count` to `max_context_source_events` config option
        :return: true if source_count within this request is equal to or
                 greater than configured threshold for source creation
        """
        if self.max_sources_logged:
            return True

        if self.exclusions and self.exclusions.evaluate_input_exclusions(
            self, source_type, source_name
        ):
            return True

        threshold_reached = self.source_count >= Settings().max_sources
        if threshold_reached:
            logger.warning(
                "Will not create more sources in this request. %s sources reached",
                self.source_count,
            )
            self.max_sources_logged = True

        return threshold_reached

    @property
    def stop_propagation(self):
        """
        Compare `propagation_count` to `max_propagation_events` config option
        :return: true if propagation_count within this request is equal to or
                 greater than configured threshold for propagation
        """
        if self.max_propagators_logged:
            return True

        threshold_reached = self.propagation_count >= Settings().max_propagation
        if threshold_reached:
            logger.warning(
                "Will not propagate any more in this request. %s propagations reached.",
                self.propagation_count,
            )
            self.max_propagators_logged = True

        return threshold_reached

    @property
    def propagate_assess(self):
        # TODO: PYT-644 move this property of out this class?
        return self.assess_enabled and not scope.in_contrast_or_propagation_scope()

    def source_created(self):
        """
        Increase the running count of sources created in this request
        """
        self.source_count += 1

    def propagated(self):
        """
        Increase the running count of propagations created in this request
        """
        self.propagation_count += 1

    def extract_response_to_context(self, response):
        """
        Append response to request context.

        :response: Subclass of BaseResponseWrapper
        """

        self.response = response

    def evaluate_exclusions(self):
        """
        Check if the request is excluded based on the URL or input exclusions.
        """
        if self.exclusions:
            path = self.request.path_info
            if self.exclusions.evaluate_url_exclusions(self, path):
                # Stop analyzing this endpoint since the URL exclusion applies
                return True

            self.exclusions.set_input_exclusions_by_url(self, path)

        return False



"""
Expose zipp.Path as .zipfile.Path
"""

import importlib
import sys
import types

from contrast_vendor import zipp


zipfile = types.SimpleNamespace(**vars(importlib.import_module('zipfile')))
zipfile.Path = zipp.Path
zipfile._path = zipp

sys.modules[__name__ + '.zipfile'] = zipfile  # type: ignore



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Route:
    """
    Route object used for various TS messages
    """

    def __init__(self, verb, url, signature, framework=None):
        self.verb = verb or "GET"
        self.url = url or "/"
        self.signature = signature or ""
        self.framework = framework
        self.sources = []

    def __hash__(self) -> int:
        return hash((self.verb, self.url, self.signature, self.framework))

    def __eq__(self, other) -> bool:
        return isinstance(other, Route) and (
            self.verb == other.verb
            and self.url == other.url
            and self.signature == other.signature
            and self.framework == other.framework
        )

    def __repr__(self):
        return f"<Route({self.verb!r}, {self.url!r}, {self.signature!r}, {self.framework!r})>"

    def to_json_inventory(self):
        """json representation used in v1.1 ApplicationInventory.routes"""
        return {
            "signature": self.signature,
            "verb": self.verb,
            "url": self.url,
            "framework": self.framework,
        }

    def to_json_traces(self):
        """json representation used in ng Traces.routes"""
        return {
            # "The number of times this route was observed; must be more than 0"
            "count": 1,
            "observations": [{"url": self.url, "verb": self.verb}],
            "signature": self.signature,
        }



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# we need this to ensure `agent_lib.modules.*` is available when importing `agent_lib`
import contrast.agent.agent_lib.modules  # noqa: F401

from contrast.agent.agent_lib.main import call, initialize, update_log_options
from contrast.agent.agent_lib.input_tracing import (
    evaluate_header_input,
    evaluate_input_by_type,
    initialize_input_tracing,
    check_sql_injection_query,
    check_cmd_injection_query,
    check_method_tampering,
    map_result_and_free_eval_result,
    map_result_and_free_check_query_sink_result,
    DBType,
)

__all__ = [
    "call",
    "initialize",
    "update_log_options",
    "evaluate_header_input",
    "evaluate_input_by_type",
    "initialize_input_tracing",
    "check_sql_injection_query",
    "check_cmd_injection_query",
    "check_method_tampering",
    "map_result_and_free_eval_result",
    "map_result_and_free_check_query_sink_result",
    "DBType",
]



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import base64
from http.cookiejar import DefaultCookiePolicy
import queue
import threading
from uuid import uuid4

from requests import Session
import contrast

from contrast.agent import scope
from contrast.agent.disable_reaction import DisableReaction
from contrast.agent.settings import Settings
from contrast.reporting.request_audit import RequestAudit
from contrast.reporting.teamserver_messages.base_ts_message import (
    PYTHON,
    b64url_stripped,
)
from contrast_vendor import structlog as logging

from .teamserver_messages import BaseTsMessage

logger = logging.getLogger("contrast")

REPORTING_CLIENT_THREAD_NAME = "ContrastReportingClient"

MAX_ATTEMPTS = 2
ERROR_STATUS_CODE = -1


class ReportingClient(threading.Thread):
    def __init__(self):
        super().__init__(name=REPORTING_CLIENT_THREAD_NAME, daemon=True)

        self.instance_id = uuid4().hex
        self._stopper = object()  # sentinel sent over message_q to stop the thread
        self.message_q = queue.Queue(maxsize=128)
        self.settings = Settings()
        self.session = Session()
        # disable persisting cookies, per Architecture team recommendation.
        self.session.cookies.set_policy(DefaultCookiePolicy(allowed_domains=[]))
        self.session.verify, self.session.cert = self.init_certs(self.settings)
        self.session.proxies = (
            self.settings.build_proxy_url() if self.settings.is_proxy_enabled else {}
        )
        server_name_b64 = b64url_stripped(self.settings.get_server_name())
        server_path_b64 = b64url_stripped(self.settings.get_server_path())
        server_type_b64 = b64url_stripped(self.settings.get_server_type())
        auth_header = f"{self.settings.api_user_name}:{self.settings.api_service_key}"
        self.session.headers.update(
            {
                # the Authorization header must not have its padding stripped
                "Authorization": base64.urlsafe_b64encode(
                    auth_header.encode()
                ).decode(),
                "API-Key": self.settings.api_key,
                "Server-Name": server_name_b64,
                "Server-Path": server_path_b64,
                "Server-Type": server_type_b64,
                "X-Contrast-Agent": f"{PYTHON} {contrast.__version__}",
                "X-Contrast-Header-Encoding": "base64",
                "X-Contrast-Reporting-Instance": self.instance_id,
                "Application-Language": PYTHON,
                "Application-Name": b64url_stripped(self.settings.app_name),
                "Application-Path": b64url_stripped(self.settings.app_path),
            }
        )

        self.request_audit = (
            RequestAudit(self.settings.config)
            if self.settings.config.is_request_audit_enabled
            else None
        )
        if self.request_audit:
            self.request_audit.prepare_dirs()

    def init_certs(self, settings):
        verify = True
        cert = None
        if not settings.is_custom_cert_enabled:
            return verify, cert

        # custom certificate settings
        if settings.ca_file:
            verify = settings.ca_file
        if settings.client_cert_file and settings.client_private_key:
            cert = (
                settings.client_cert_file,
                settings.client_private_key,
            )
        if settings.ignore_cert_errors:
            verify = False
            logger.warning("Certificate verification is disabled.")

        # misconfig checks
        if (settings.client_cert_file and not settings.client_private_key) or (
            not settings.client_cert_file and settings.client_private_key
        ):
            logger.error(
                "Unable to communicate with Contrast. "
                "Certificate configuration is not set properly. "
                "Certificate PEM file or private key PEM file is missing.",
                cert_file=settings.client_cert_file,
                key_file=settings.client_private_key,
            )
            DisableReaction.run(settings)
            return verify, cert

        if not any(
            (settings.ca_file, settings.client_cert_file, settings.client_private_key)
        ):
            logger.warning(
                "Certificate configuration is enabled, but no certificate files are set."
            )

        return verify, cert

    def stop(self):
        self.message_q.put(self._stopper)
        self.join()

    def run(self):
        with scope.contrast_scope():
            logger.debug("Starting reporting thread")

            while self.settings.is_agent_config_enabled():
                try:
                    msg = self.message_q.get(block=True, timeout=5)
                    if msg is self._stopper:
                        break
                    response = self.send_message(msg)
                    msg.process_response(response, self)
                except queue.Empty:
                    pass
                except Exception as e:
                    logger.debug(
                        "WARNING: reporting client failed to send message", exc_info=e
                    )

    def send_message(self, msg):
        try:
            return self._send_message(msg)
        except Exception as e:
            logger.exception(
                "Failed to send %s message to Contrast: %s", msg.class_name, e
            )

        return None

    def _send_message(self, msg: BaseTsMessage):
        """
        _send_message sends msg to Teamserver and returns the response.

        It is the caller's responsibility to handle any network exceptions.

        See send_message for the public interface, which handles exceptions.
        """
        status_code = ERROR_STATUS_CODE
        msg_name = msg.class_name

        url = msg.base_url + msg.path
        logger.debug("Sending %s message to Teamserver: %s", msg_name, url)
        response = msg.request_method(
            self.session,
            url,
            json=msg.body,
            headers=msg.headers,
            allow_redirects=False,
            # environment variables can interfere with session fields, so
            # we set them explicitly as arguments:
            # https://github.com/psf/requests/issues/3829
            proxies=self.session.proxies,
            verify=self.session.verify,
            cert=self.session.cert,
        )

        try:
            status_code = response.status_code
            msg_success_status = response.json().get("success")
            messages = response.json().get("messages")
            if not (msg_success_status or status_code == 200):
                logger.error(
                    "Failure on Contrast UI processing request reason - (%s): %s",
                    messages,
                    status_code,
                )
        except Exception as e:
            if status_code == ERROR_STATUS_CODE:
                logger.debug(
                    "Failed to receive response from Contrast UI: %s ",
                    e,
                )

        logger.debug("Contrast UI response (%s): %s", msg_name, status_code)

        if self.request_audit:
            self.request_audit.audit(msg, response)

        msg.sent()
        return response

    def add_message(self, msg):
        if msg is None or not isinstance(msg, BaseTsMessage):
            return

        logger.debug("Adding msg to reporting queue: %s", msg.class_name)

        self.message_q.put(msg)

    def retry_message(self, msg):
        # Never send a message more than twice (original time plus one retry)
        # To prevent queue from filling up or causing memory issues.
        if msg.sent_count < MAX_ATTEMPTS:
            logger.debug("Re-enqueuing %s message", msg.class_name)
            self.add_message(msg)



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .finding import Finding, send_finding
from .library import Library
from .route_coverage import Route
from .type_checked_property import TypeCheckedProperty

__all__ = [
    "Finding",
    "send_finding",
    "Library",
    "Route",
    "TypeCheckedProperty",
]



# Copyright © 2024 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Literal
import requests

from .base_ts_message import BaseTsServerMessage
from contrast.utils.decorators import fail_loudly


class ServerInventory(BaseTsServerMessage):
    def __init__(self, cloud_provider: Literal["aws", "azure"], cloud_resource_id: str):
        super().__init__()

        # TODO: PYT-3378 use v1.1
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

        self.body = {
            "cloud_provider": cloud_provider,
            "cloud_resource_id": cloud_resource_id,
        }

    # TODO: PYT-3381 Remove special 404 logic completely for this message
    @property
    def ignore_404(self) -> bool:
        return True

    @property
    def name(self) -> str:
        return "server-inventory"

    @property
    def path(self) -> str:
        return "/".join(
            [
                "servers",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                "inventory",
            ]
        )

    @property
    def request_method(self):
        return requests.Session.post

    @fail_loudly()
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)
